#!/usr/bin/env python3
import os, os.path, sys
from .argparse_compat import argparse
from .import_util import get_file, sort_files_by_dependency, IMPORT_ABSOLUTIZE_TUPLE, ALL_ABSOLUTIZE_TUPLE
from .custom_arguments import (
    add_libname_arguments,
    update_env_with_libnames,
    add_logging_arguments,
    process_logging_arguments,
    get_parser_name_mapping,
)
from .file_util import write_to_file

__all__ = ["main"]

parser = argparse.ArgumentParser(description="Absolutize the imports of Coq files")
parser.add_argument("input_files", metavar="INFILE", nargs="*", type=argparse.FileType("r"), help=".v files to update")
parser.add_argument(
    "--in-place",
    "-i",
    metavar="SUFFIX",
    dest="suffix",
    nargs="?",
    type=str,
    default="",
    help="update files in place (makes backup if SUFFIX supplied)",
)
parser.add_argument(
    "--update-all",
    dest="update_all",
    action="store_const",
    default=False,
    const=True,
    help=("also update all .v files listed in any _CoqProject file passed to -f (implies --in-place, requires -f)"),
)
parser.add_argument(
    "--all",
    "-a",
    dest="absolutize",
    action="store_const",
    const=ALL_ABSOLUTIZE_TUPLE,
    default=IMPORT_ABSOLUTIZE_TUPLE,
    help="Absolutize all constants, and not just imports.",
)
parser.add_argument(
    "--coqbin",
    metavar="COQBIN",
    dest="coqbin",
    type=str,
    default="",
    help="The path to a folder containing the coqc and coqtop programs.",
)
parser.add_argument(
    "--coqc", metavar="COQC", dest="coqc", type=str, default=None, action="append", help="The path to the coqc program."
)
parser.add_argument(
    "--coq_makefile",
    metavar="COQ_MAKEFILE",
    dest="coq_makefile",
    type=str,
    default=None,
    action="append",
    help="The path to the coq_makefile program.",
)
add_libname_arguments(parser)
add_logging_arguments(parser)


def absolutize_imports(filename, **kwargs):
    kwargs["log"]("Processing %s..." % filename)
    absolutized_contents = get_file(filename, update_globs=True, **kwargs)
    if kwargs["inplace"]:
        do_backup = kwargs["suffix"] is not None and len(kwargs["suffix"]) > 0
        write_to_file(filename, absolutized_contents, do_backup=do_backup, backup_ext=kwargs["suffix"])
    else:
        print(absolutized_contents)


def main():
    args = process_logging_arguments(parser.parse_args())

    def prepend_coqbin(prog):
        if isinstance(prog, str):
            prog = (prog,)
        if args.coqbin != "":
            return (os.path.join(args.coqbin, prog[0]), *prog[1:])
        else:
            return tuple(prog)

    env = {
        "log": args.log,
        "coqc": prepend_coqbin(args.coqc or "coqc"),
        "inplace": args.suffix != "",  # it's None if they passed no argument, and '' if they didn't pass -i
        "suffix": args.suffix,
        "absolutize": args.absolutize,
        "coq_makefile": prepend_coqbin(args.coq_makefile or "coq_makefile"),
        "input_files": tuple(f.name for f in args.input_files),
        "cli_mapping": get_parser_name_mapping(parser),
    }
    update_env_with_libnames(env, args)

    for f in args.input_files:
        f.close()

    if args.update_all:
        if env["_CoqProject"] is None:
            parser.error("--update-all given without -f")
            sys.exit(1)
        else:
            env["inplace"] = True
            if env["suffix"] == "":
                env["suffix"] = None
            env["input_files"] = tuple(list(env["input_files"]) + list(env["_CoqProject_v_files"]))
        if len(env["input_files"]) == 0:
            parser.error("no .v files listed in %s" % args.CoqProjectFile)
    elif len(env["input_files"]) == 0:
        parser.error("not enough arguments (-f COQPROJECTFILE with .v files is required if no .v files are given)")

    env["input_files"] = sort_files_by_dependency(env["input_files"], update_globs=True, **env)

    for f in env["input_files"]:
        absolutize_imports(f, **env)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from __future__ import with_statement
import re
from .argparse_compat import argparse
from .split_file import split_coq_file_contents_with_comments
from .strip_comments import strip_comments
from .custom_arguments import add_logging_arguments, process_logging_arguments, get_parser_name_mapping, LOG_ALWAYS
from .file_util import write_to_file
from .util import PY3

if PY3:
    from .util import raw_input

__all__ = ["main"]

parser = argparse.ArgumentParser(description="Move various statements out of proof blocks")
parser.add_argument("input_files", metavar="INFILE", nargs="+", type=str, help=".v files to update")
parser.add_argument(
    "--in-place",
    "-i",
    metavar="SUFFIX",
    dest="suffix",
    nargs="?",
    type=str,
    default="",
    help="update files in place (makes backup if SUFFIX supplied)",
)
add_logging_arguments(parser)

ALL_DEFINITIONS_REG = re.compile(
    r"^\s*(?:(?:Global|Local|Polymorphic|Monomorphic|Time|Timeout)\s+)?(?:"
    + r"Theorem|Lemma|Fact|Remark|Corollary|Proposition|Property"
    + r"|Definition|Example|SubClass"
    + r"|Let|Fixpoint|CoFixpoint"
    + r"|Structure|Coercion|Instance"
    + r"|Add Parametric Morphism"
    + r")\s",
    re.MULTILINE,
)

ONELINE_DEFINITIONS_REG = re.compile(
    r"^\s*(?:(?:Global|Local|Polymorphic|Monomorphic|Time|Timeout)\s+)?(?:" + r"Coercion|Existing\s+Instance" + r")\s",
    re.MULTILINE,
)

ALL_ENDINGS = re.compile(r"^(?:[}{]|\s)*(?:(?:Time|Timeout)\s+)?(?:Qed|Defined|Save|Admitted|Abort)\s*\.", re.MULTILINE)

COPY_UP_REG = re.compile(
    r"^\s*(?:(?:Global|Local|Polymorphic|Monomorphic|Time|Timeout)\s+)?"
    + r"(?:Opaque\s|Transparent\s|Arguments\s|Implicit\s+Arguments\s)",
    re.MULTILINE,
)
MOVE_UP_REG = re.compile(
    r"^\s*(?:(?:Global|Local|Polymorphic|Monomorphic|Time|Timeout)\s+)?"
    + r"(?:Require|Import|Notation|Ltac|Tactic\s+Notation|Infix|Delimit\s+Scope|Reserved\s+Notation|Reserved\+Infix|Existing\s+Instance|Coercion|Hint|Set\s+Printing|Unset\+Printing)\s",
    re.MULTILINE,
)
SPECIAL_TACTICS = r"W_eq|PropXRel|ILAlgoTypes|NoDup|W_neq|PropXTac|SEP_FACTS|SF\."
SAFE_REG = re.compile(
    r"^\s*(?:(?:Global|Local|Polymorphic|Monomorphic|Time|Timeout)\s+)?(?:"
    + r"[a-z]|[0-9]+\s*:|\(\s*[a-z]|"
    + SPECIAL_TACTICS
    + r"|{|}|-|\+|\*"
    + r"|(?:Unfocus|Proof|Focus|Grab\s+Existentials?|Open\s+Scope|Close\s+Scope)(?:\.|\s)|\(\*"
    + r")",
    re.MULTILINE,
)

reg = re.compile(r"Require\s.*?\.\s+", re.MULTILINE | re.DOTALL)


def get_leading_space(string):
    return re.findall(r"^(?:\s*?\n)?([ \t]*)(?!\s)", string)[0]


def remove_leading_space(string, space_count):
    return re.sub(r"(^|\n)" + (" " * space_count), r"\1", string, re.MULTILINE)


def set_leading_space(string, space_count):
    return remove_leading_space(string, max(0, len(get_leading_space(string)) - space_count))


def strip_parens(string):
    last = string
    cur = re.sub('"[^"]+|{[^{}]+}|\([^\(\)]+\)', "", last)
    while cur != last:
        last, cur = cur, re.sub('"[^"]+|{[^{}]+}|\([^\(\)]+\)', "", cur)
    cur = re.sub(r'\slet\s[^\(\){}"]+?:=[^\(\){}"]+?\sin\s', "", cur, re.MULTILINE)
    return cur


def space_canonicalize(stmt):
    return re.sub(r"\s+", " ", stmt.strip(" \t\n\r."))


def canonicalize(stmt):
    return space_canonicalize(stmt).replace("Transparent ", "Opaque ")


def cancels(statement1, statement2):
    """Returns True if statement2 cancels the effect of statement1; False otherwise"""
    return canonicalize(statement1) == canonicalize(statement2)


def commutes(statement1, statement2):
    if any(i in statement1 or i in statement2 for i in ("(*", "*)", '"')):
        return False
    statement1, statement2 = space_canonicalize(statement1), space_canonicalize(statement2)
    m1 = re.match(r"^(?:Local |Global )?(?:Transparent|Opaque) (.*)$", statement1)
    m2 = re.match(r"^(?:Local |Global )?(?:Transparent|Opaque) (.*)$", statement2)
    if not m1 or not m2:
        return False
    constants1, constants2 = set(m1.groups()[0].split(" ")), set(m2.groups()[0].split(" "))
    return constants1.isdisjoint(constants2)


def preminimize_lifted_statements(statements):
    """Remove useless deferred statements, such as transparent followed by opaque"""

    prev = set()
    for statement in statements:
        prev = set(last for last in prev if not cancels(last, statement))
        if not all(commutes(last, statement) for last in prev):
            for last in sorted(prev):
                yield last
            prev = set()
        prev.add(statement)
    for last in sorted(prev):
        yield last


def minimize_lifted_statements(statements):
    statements = list(preminimize_lifted_statements(statements))
    sans_copy = [i for i in statements if COPY_UP_REG.match(i.strip()) is None]
    sans_copy_hint = [i for i in sans_copy if re.match(r"^(?:Local\s+|Global\s+)?Hint\s", i.strip()) is None]
    if len(sans_copy_hint) == 1:
        return sans_copy
    else:
        return statements


def move_from_proof(filename, **kwargs):
    kwargs["log"]("Processing %s..." % filename)
    try:
        with open(filename, "r") as f:
            contents = f.read()
    except IOError as e:
        kwargs["log"]("Failed to process %s" % filename)
        kwargs["log"](repr(e), level=2)
        return
    ret = []
    cur_statements = []
    cur_statement = []
    deferred_copied_statements = []
    deferred_statements = []
    orig_space_count = 0
    cur_diff_space_count = 0
    if "".join(split_coq_file_contents_with_comments(contents)) != contents:
        kwargs["log"]("WARNING: Could not split %s" % filename, level=LOG_ALWAYS)
        return
    for i in split_coq_file_contents_with_comments(contents):
        is_definition_full = ALL_DEFINITIONS_REG.match(i) is not None and (
            ":=" in strip_parens(strip_comments(i)) or ONELINE_DEFINITIONS_REG.match(i)
        )
        is_definition_start = (
            ALL_DEFINITIONS_REG.match(i) is not None
            and ":=" not in strip_parens(strip_comments(i))
            and not ONELINE_DEFINITIONS_REG.match(i)
        )
        # print((is_definition_start, ONELINE_DEFINITIONS_REG.match(i), ALL_DEFINITIONS_REG.match(i), i))
        is_definition_end = ALL_ENDINGS.match(i) is not None
        # if 'not_reachable_iff' in i:
        #    print((ALL_DEFINITIONS_REG.match(i), strip_parens(strip_comments(i)), ONELINE_DEFINITIONS_REG.match(i)))
        if not is_definition_start and not cur_statements and not cur_statement:
            kwargs["log"](repr(i), level=3)
            ret.append(i)
        elif is_definition_start:
            kwargs["log"]("Starting definition (%d): %s" % (len(deferred_statements), repr(i)), level=2)
            if not cur_statement and not deferred_statements:
                orig_space_count = len(get_leading_space(i))
            if cur_statement:
                deferred_statements.append((cur_diff_space_count, cur_statement))
            cur_diff_space_count = max(0, len(get_leading_space(i)) - orig_space_count)
            cur_statement = [remove_leading_space(i, cur_diff_space_count)]
        elif (SAFE_REG.match(i) or not i.strip()) and cur_statement:
            kwargs["log"](repr(i), level=3)
            cur_statement.append(remove_leading_space(i, cur_diff_space_count))
        elif is_definition_end and cur_statement:
            kwargs["log"]("Ending definition: " + repr(i), level=2)
            cur_statement.append(remove_leading_space(i, cur_diff_space_count))
            cur_statements.append("".join(cur_statement))
            # print(''.join(preminimize_lifted_statements(cur_statements)))
            if deferred_statements:
                cur_diff_space_count, cur_statement = deferred_statements.pop()
                deferred_copied_statements = list(preminimize_lifted_statements(deferred_copied_statements))
                cur_statement.extend(deferred_copied_statements)  # todo: fix indentation
            else:
                # print('extending')
                # print(''.join(minimize_lifted_statements(cur_statements)))
                ret.extend(list(minimize_lifted_statements(cur_statements)))
                cur_statement = []
                cur_statements = []
                deferred_copied_statements = []
        elif MOVE_UP_REG.match(i) or is_definition_full:
            kwargs["log"]("Lifting: " + repr(i), level=2)
            cur_statements.append(set_leading_space(i, orig_space_count))
        elif COPY_UP_REG.match(i) and cur_statement:
            kwargs["log"]("Lift-copying: " + repr(i), level=2)
            cur_statement.append(remove_leading_space(i, cur_diff_space_count))
            cur_statements.append(set_leading_space(i, orig_space_count))
            deferred_copied_statements.append(remove_leading_space(i, cur_diff_space_count))
        else:
            raw_input("WARNING: Unrecognized: %s" % repr(i))
        # print(cur_diff_space_count)
        # print(remove_leading_space(i, cur_diff_space_count))
    if cur_statements or deferred_statements or cur_statement:
        raw_input("WARNING: extra statements: %s" % repr((cur_statements, cur_statement, deferred_statements)))
        cur_statements.append("".join(cur_statement))
        for i in deferred_statements:
            cur_statements.append("".join(i))
        ret.extend(list(minimize_lifted_statements(cur_statements)))
    ret = "".join(ret)
    if ret == contents:
        return
    if kwargs["inplace"]:
        do_backup = kwargs["suffix"] is not None and len(kwargs["suffix"]) > 0
        write_to_file(filename, ret, do_backup=do_backup, backup_ext=kwargs["suffix"])
    else:
        print(ret)


def main():
    args = process_logging_arguments(parser.parse_args())
    env = {
        "log": args.log,
        "inplace": args.suffix != "",  # it's None if they passed no argument, and '' if they didn't pass -i
        "suffix": args.suffix,
        "cli_mapping": get_parser_name_mapping(parser),
    }

    for f in args.input_files:
        move_from_proof(f, **env)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from __future__ import with_statement
import os, sys, re
from .argparse_compat import argparse
from .custom_arguments import get_parser_name_mapping

__all__ = ["main"]

parser = argparse.ArgumentParser(description="Move the [Require] statements up to the top of a file")
parser.add_argument("input_files", metavar="INFILE", nargs="+", type=str, help=".v files to update")
parser.add_argument(
    "--in-place",
    "-i",
    metavar="SUFFIX",
    dest="suffix",
    nargs="?",
    type=str,
    default="",
    help="update files in place (makes backup if SUFFIX supplied)",
)
parser.add_argument("--verbose", "-v", dest="verbose", action="count", help="display some extra information")
parser.add_argument("--quiet", "-q", dest="quiet", action="count", help="the inverse of --verbose")
parser.add_argument(
    "--log-file",
    "-l",
    dest="log_files",
    nargs="*",
    type=argparse.FileType("w"),
    default=[sys.stderr],
    help="The files to log output to.  Use - for stdout.",
)

DEFAULT_VERBOSITY = 1


def make_logger(log_files):
    def log(text):
        for i in log_files:
            i.write(str(text) + "\n")
            i.flush()
            if i.fileno() > 2:  # stderr
                os.fsync(i.fileno())

    return log


DEFAULT_LOG = make_logger([sys.stderr])


def backup(file_name, ext=".bak"):
    if not ext:
        raise ValueError
    if os.path.exists(file_name):
        backup(file_name + ext)
        os.rename(file_name, file_name + ext)


def write_to_file(file_name, contents, do_backup=False, ext=".bak"):
    if do_backup:
        backup(file_name, ext=ext)
    try:
        with open(file_name, "w", encoding="UTF-8") as f:
            f.write(contents)
    except TypeError:
        with open(file_name, "w") as f:
            f.write(contents)


reg = re.compile(r"Require\s.*?\.\s+", re.MULTILINE | re.DOTALL)


def move_requires(filename, **kwargs):
    if kwargs["verbose"]:
        kwargs["log"]("Processing %s..." % filename)
    with open(filename, "r") as f:
        contents = f.read()
    requires = reg.findall(contents)
    if len(requires) == 0:
        if "Require" not in contents:
            return
        else:
            kwargs["log"]("Warning: Inconsistent [Require] in %s" % filename)
    header_end = contents.index(requires[0])
    header, rest = contents[:header_end], contents[header_end:]
    rest = reg.sub("", rest)
    new_contents = header + "\n".join(map(str.strip, requires)) + "\n\n" + rest
    if kwargs["inplace"]:
        do_backup = kwargs["suffix"] is not None and len(kwargs["suffix"]) > 0
        write_to_file(filename, new_contents, do_backup=do_backup, ext=kwargs["suffix"])
    else:
        print(new_contents)


def main():
    args = parser.parse_args()
    if args.verbose is None:
        args.verbose = DEFAULT_VERBOSITY
    if args.quiet is None:
        args.quiet = 0
    verbose = args.verbose - args.quiet
    log = make_logger(args.log_files)
    env = {
        "verbose": verbose,
        "log": log,
        "inplace": args.suffix != "",  # it's None if they passed no argument, and '' if they didn't pass -i
        "suffix": args.suffix,
        "cli_mapping": get_parser_name_mapping(parser),
    }

    for f in args.input_files:
        move_requires(f, **env)


if __name__ == "__main__":
    main()



__all__ = [
    "absolutize-imports",
    "find_bug",
    "get-admitted-names",
    "inline_imports",
    "minimize_requires",
    "move-requires",
    "move-vernaculars",
    "proof-using-helper",
]



import sys

if sys.version_info < (3,):
    from . import argparse_py2 as argparse
else:
    import argparse

if not hasattr(argparse, "BooleanOptionalAction"):
    # from the stdlib

    class BooleanOptionalAction(argparse.Action):
        def __init__(
            self, option_strings, dest, default=None, type=None, choices=None, required=False, help=None, metavar=None
        ):

            _option_strings = []
            for option_string in option_strings:
                _option_strings.append(option_string)

                if option_string.startswith("--"):
                    option_string = "--no-" + option_string[2:]
                    _option_strings.append(option_string)

            super().__init__(
                option_strings=_option_strings,
                dest=dest,
                nargs=0,
                default=default,
                type=type,
                choices=choices,
                required=required,
                help=help,
                metavar=metavar,
            )

        def __call__(self, parser, namespace, values, option_string=None):
            if option_string in self.option_strings:
                setattr(namespace, self.dest, not option_string.startswith("--no-"))

        def format_usage(self):
            return " | ".join(self.option_strings)

    argparse.BooleanOptionalAction = BooleanOptionalAction



from __future__ import with_statement, print_function
import re
from .split_file import split_leading_comments_and_whitespace
from .import_util import (
    filename_of_lib,
    lib_of_filename,
    get_file,
    run_recursively_get_imports,
    run_maybe_recursively_get_imports,
    recursively_get_imports,
    absolutize_has_all_constants,
    is_local_import,
    ALL_ABSOLUTIZE_TUPLE,
)
from .custom_arguments import DEFAULT_LOG
from .coq_running_support import get_reserved_modnames

__all__ = [
    "include_imports",
    "normalize_requires",
    "get_required_contents",
    "recursively_get_requires_from_file",
    "absolutize_and_mangle_libname",
]

file_contents = {}
lib_imports_fast = {}
lib_imports_slow = {}

DEFAULT_LIBNAMES = ((".", "Top"),)

IMPORT_LINE_REG = re.compile(
    r"^\s*(?:From|Require\s+Import|Require\s+Export|Require|Load\s+Verbose|Load)\s+(.*?)\.(?:\s|$)",
    re.MULTILINE | re.DOTALL,
)


def fill_kwargs(kwargs, for_makefile=True):
    rtn = {
        "fast": False,
        "log": DEFAULT_LOG,
        "libnames": DEFAULT_LIBNAMES,
        "non_recursive_libnames": (),
        "coqc": ("coqc",),
        "coqtop": ("coqtop",),
        "absolutize": ALL_ABSOLUTIZE_TUPLE,
        "coq_makefile": ("coq_makefile",),
    }
    rtn.update(kwargs)
    if for_makefile:
        if "make_coqc" in rtn.keys():  # handle the case where coqc for the makefile is different
            rtn["coqc"] = rtn["make_coqc"]
        if "passing_make_coqc" in rtn.keys():  # handle the case where coqc for the makefile is different
            rtn["passing_coqc"] = rtn["passing_make_coqc"]
    return rtn


def contents_without_imports(lib, **kwargs):
    v_file = filename_of_lib(lib, ext=".v", **kwargs)
    contents = get_file(v_file, **kwargs)
    if "(*" in " ".join(IMPORT_LINE_REG.findall(contents)):
        print("Warning: There are comments in your Require/Import/Export lines in %s." % v_file)
    return IMPORT_LINE_REG.sub("", contents)


def escape_lib(lib, **kwargs):
    reserved = get_reserved_modnames(kwargs["coqtop"], **kwargs)
    lib = lib.replace(".", "_DOT_").replace("-", "_DASH_")
    while lib in reserved:
        lib = "AVOID_RESERVED_" + lib
    return lib


def group_by_first_component(lib_libname_pairs):
    rtn = dict((lib.split(".")[0], []) for lib, libname in lib_libname_pairs)
    for lib, libname in lib_libname_pairs:
        split_lib = lib.split(".")
        rtn[split_lib[0]].append((".".join(split_lib[1:]), libname))
    return rtn


def nest_iter_up_to(iterator):
    so_far = []
    for i in iterator:
        so_far.append(i)
        yield tuple(so_far)


def construct_import_list(import_libs, import_all_directories=False, **kwargs):
    """Takes a list of library names, and returns a list of imports in an order that should have modules representing files at the end.  If import_all_directories is true, then the resulting imports should handle semi-absolute constants, and not just fully absolute or fully relative ones."""

    def escape_lib_local(l):
        return escape_lib(l, **kwargs)

    if import_all_directories:
        lib_components_list = [
            (libname, tuple(reversed(list(nest_iter_up_to(map(escape_lib_local, libname.split("."))))[:-1])))
            for libname in import_libs
        ]
        ret = list(map(escape_lib_local, import_libs))
        lib_components = [
            (libname, i, max(map(len, lst)) - len(i)) for libname, lst in lib_components_list for i in lst
        ]
        for libname, components, components_left in reversed(sorted(lib_components, key=(lambda x: x[2]))):
            ret.append(escape_lib_local(libname) + "." + ".".join(map(escape_lib_local, components)))
        return ret
    else:
        return map(escape_lib_local, import_libs)


def strip_requires(contents):
    reg1 = re.compile(r"^\s*Require\s+((?:Import|Export)\s)", flags=re.MULTILINE)
    contents = reg1.sub(r"\1", contents)
    reg2 = re.compile(r"^\s*Require\s+((?!Import\s+|Export\s+)(?:[^\.]|\.(?!\s|$))+\.(?:\s|$))", flags=re.MULTILINE)
    contents = reg2.sub(r"", contents)
    reg3 = re.compile(r"^\s*(From\s+[^\s]+\s+)Require\s+((?:Import|Export)\s)", flags=re.MULTILINE)
    contents = reg3.sub(r"\1\2", contents)
    reg4 = re.compile(
        r"^\s*(From\s+[^\s]+\s+)Require\s+((?!Import\s+|Export\s+)(?:[^\.]|\.(?!\s|$))+\.(?:\s|$))", flags=re.MULTILINE
    )
    contents = reg4.sub(r"", contents)
    return contents


def get_module_name_and_lib_parts(lib, first_wrap_then_include=False, **kwargs):
    def escape_lib_local(libname):
        return escape_lib(libname, **kwargs)

    module_name = escape_lib_local(lib)
    mangled_module_name = module_name + "_WRAPPED"
    lib_parts = list(map(escape_lib_local, lib.split(".")))
    # we could actually return the same thing, that is the string
    # ('%s.%s' % (module_name, '.'.join(lib_parts))), in both cases,
    # but we choose to return the module prior to Include rather than
    # the wrapped module, when we're wrapping via include, so that we
    # have a chance of removing the Include statement later (because
    # it may not be used)
    if first_wrap_then_include:
        full_module_name = "%s.%s" % (mangled_module_name, lib_parts[-1])
    else:
        full_module_name = "%s.%s" % (module_name, ".".join(lib_parts))
    return module_name, mangled_module_name, lib_parts, full_module_name


def absolutize_and_mangle_libname(lib, first_wrap_then_include=False, **kwargs):
    module_name, mangled_module_name, lib_parts, full_module_name = get_module_name_and_lib_parts(
        lib, first_wrap_then_include=first_wrap_then_include, **kwargs
    )
    return full_module_name


def contents_as_module(
    lib,
    other_imports,
    extra_contents_inside_module: str = "",
    first_wrap_then_include=False,
    export=False,
    without_require=True,
    extra_top_header=None,
    **kwargs,
):
    import_all_directories = not absolutize_has_all_constants(kwargs["absolutize"])
    if import_all_directories and not export:
        # N.B. This strategy does not work with the Include strategy
        # that we use to fix
        # https://github.com/JasonGross/coq-tools/issues/67, so we disable it
        first_wrap_then_include = False
        transform_base = lambda x: (escape_lib(x, **kwargs) + "." + x if is_local_import(x, **kwargs) else x)
    else:
        transform_base = lambda x: x
    v_name = filename_of_lib(lib, ext=".v", **kwargs)
    contents = get_file(v_name, transform_base=transform_base, **kwargs)
    if without_require:
        contents = strip_requires(contents)
    kwargs["log"](contents, level=3)
    module_name, mangled_module_name, lib_parts, _ = get_module_name_and_lib_parts(
        lib, first_wrap_then_include=first_wrap_then_include, **kwargs
    )
    # import the top-level wrappers
    if len(other_imports) > 0 and not export:
        # we need to import the contents in the correct order.  Namely, if we have a module whose name is also the name of a directory (in the same folder), we want to import the file first.
        for imp in reversed(
            construct_import_list(other_imports, import_all_directories=import_all_directories, **kwargs)
        ):
            contents = "Import %s.\n%s" % (imp, contents)
    # insert extra_top_header, e.g., Import Coq.Init.Prelude., at the top of the file
    if extra_top_header:
        contents = extra_top_header + "\n" + contents
    # wrap the contents in directory modules
    maybe_export = "Export " if export else ""
    if extra_contents_inside_module:
        contents = extra_contents_inside_module + "\n" + contents
    early_contents = ""
    if first_wrap_then_include:  # works around https://github.com/JasonGross/coq-tools/issues/67
        early_contents, contents = contents, "Include %s.%s." % (mangled_module_name, lib_parts[-1])
        early_contents = "Module %s.\n%s\nEnd %s.\n" % (lib_parts[-1], early_contents, lib_parts[-1])
        early_contents = "Module %s.\n%s\nEnd %s.\n" % (mangled_module_name, early_contents, mangled_module_name)
    contents = "Module %s.\n%s\nEnd %s.\n" % (lib_parts[-1], contents, lib_parts[-1])
    for name in reversed(lib_parts[:-1]):
        contents = "Module %s%s.\n%s\nEnd %s.\n" % (maybe_export, name, contents, name)
    contents = "Module %s%s.\n%s\nEnd %s.\n" % (maybe_export, module_name, contents, module_name)
    return early_contents + contents


def normalize_requires(filename, recursive_requires_explicit: bool = True, **kwargs):
    """Return the contents of filename, with all [Require]s split out and ordered at the top.

    Preserve any leading whitespace/comments.
    """
    if filename[-2:] != ".v":
        filename += ".v"
    kwargs = fill_kwargs(kwargs)
    lib = lib_of_filename(filename, **kwargs)
    all_imports = run_maybe_recursively_get_imports(lib, recursively=recursive_requires_explicit, **kwargs)

    v_name = filename_of_lib(lib, ext=".v", **kwargs)
    contents = get_file(v_name, **kwargs)
    header, contents = split_leading_comments_and_whitespace(contents)
    contents = strip_requires(contents)
    contents = "".join("Require %s.\n" % i for i in all_imports[:-1]) + "\n" + contents.strip() + "\n"
    return header + contents


def get_required_contents(libname, **kwargs):
    return contents_as_module(libname, other_imports=[], export=True, **fill_kwargs(kwargs))


def recursively_get_requires_from_file(filename, **kwargs):
    if filename[-2:] != ".v":
        filename += ".v"
    kwargs = fill_kwargs(kwargs)
    lib = lib_of_filename(filename, **kwargs)
    return tuple(run_recursively_get_imports(lib, **kwargs)[:-1])


def include_imports(filename, as_modules=True, absolutize=ALL_ABSOLUTIZE_TUPLE, **kwargs):
    """Return the contents of filename, with any top-level imports inlined.

    If as_modules == True, then the imports will be wrapped in modules.

    This method requires access to the coqdep program if fast == False.
    Without it, it will fall back to manual parsing of the imports,
    which may change behavior.

    >>> import tempfile, os
    >>> f = tempfile.NamedTemporaryFile(dir='.', suffix='.v', delete=False)
    >>> g = tempfile.NamedTemporaryFile(dir='.', suffix='.v', delete=False)
    >>> g_name = os.path.relpath(g.name, '.')[:-2]
    >>> f.write("  Require  Import %s Coq.Init.Logic Arith.\\n  Require  Export \tPArith\t Coq.Init.Logic.\\n  Require Bool.\\n Import Bool. (* asdf *)\\n Require Import QArith\\n  ZArith\\n  Setoid.\\nRequire Import %s.\\n Require\\n  Import\\n%s\\n\\n\t.\t(*foo*)\\n\\nInductive t := a | b.\\n\\n(*asdf*)" % (g_name, g_name, g_name))
    >>> g.write(r"Require Export Ascii String.\\n\\nInductive q := c | d.")
    >>> f.close()
    >>> g.close()
    >>> print(include_imports(f.name, as_modules=False))
    Require Import Coq.Arith.Arith Coq.Bool.Bool Coq.Init.Logic Coq.PArith.PArith Coq.QArith.QArith Coq.Setoids.Setoid Coq.ZArith.ZArith Coq.Strings.Ascii Coq.Strings.String.

    Inductive q := c | d.
    (* asdf *)


    Inductive t := a | b.

    (*asdf*)

    >>> print(include_imports(f.name, as_modules=False, fast=True))
    Require Import Arith Bool Coq.Init.Logic PArith QArith Setoid ZArith Ascii String.

    Inductive q := c | d.
    (* asdf *)


    Inductive t := a | b.

    (*asdf*)
    >>> exts = ('.v', '.v.d', '.glob', '.vo', '.o', '.cmi', '.cmxs', '.native', '.cmx', '.vok', '.vos')
    >>> names = [f.name[:-2] + ext for ext in exts] + [g.name[:-2] + ext for ext in exts]
    >>> names = [i for i in names if os.path.exists(i)]
    >>> for name in names: os.remove(name)
    """
    kwargs = fill_kwargs(kwargs)
    del kwargs["absolutize"]
    if "make_coqc" in kwargs.keys():
        kwargs["coqc"] = kwargs["make_coqc"]
    if filename[-2:] != ".v":
        filename += ".v"
    lib = lib_of_filename(filename, **kwargs)
    all_imports = recursively_get_imports(lib, **kwargs)
    remaining_imports = []
    rtn = ""
    imports_done = []
    for import_name in all_imports:
        try:
            if as_modules:
                rtn += (
                    contents_as_module(
                        import_name,
                        imports_done,
                        absolutize=absolutize,
                        without_require=True,
                        **kwargs,
                    )
                    + "\n"
                )
            else:
                rtn += contents_without_imports(import_name, absolutize=(), **kwargs) + "\n"
            imports_done.append(import_name)
        except IOError:
            remaining_imports.append(import_name)
    if len(remaining_imports) > 0:
        kwargs["log"]("remaining imports:", level=2)
        kwargs["log"](remaining_imports, level=2)
        if as_modules:
            pattern = "Require %s.\n%s"
        else:
            pattern = "Require Import %s.\n%s"
        for imp in reversed(remaining_imports):
            rtn = pattern % (imp, rtn)
    return rtn


if __name__ == "__main__":
    # if we're working in Python 3.3, we can test this file
    try:
        import doctest

        success = True
    except ImportError:
        print(
            "This is not the main file to use.\nOnly run it if you have doctest (Python 3.3+) and are testing things."
        )
        success = False
    if success:
        doctest.testmod()



from __future__ import print_function
import sys, os
from .argparse_compat import argparse
from .util import PY3

__all__ = [
    "add_libname_arguments",
    "add_passing_libname_arguments",
    "ArgumentParser",
    "update_env_with_libnames",
    "add_logging_arguments",
    "process_logging_arguments",
    "update_env_with_coqpath_folders",
    "DEFAULT_LOG",
    "DEFAULT_VERBOSITY",
    "LOG_ALWAYS",
    "get_parser_name_mapping",
]


# grumble, grumble, we want to support multiple -R arguments like coqc
class CoqLibnameAction(argparse.Action):
    non_default = False

    #     def __init__(self, *args, **kwargs):
    #         super(CoqLibnameAction, self).__init__(*args, **kwargs)
    def __call__(self, parser, namespace, values, option_string=None):
        #        print('%r %r %r' % (namespace, values, option_string))
        if not self.non_default:
            setattr(namespace, self.dest, [])
            self.non_default = True
        getattr(namespace, self.dest).append(tuple(values))


DEFAULT_VERBOSITY = 1
LOG_ALWAYS = None


def make_logger(log_files):
    # log if level <= the level of the logger
    def log(
        text,
        level=DEFAULT_VERBOSITY,
        max_level=None,
        force_stdout=False,
        force_stderr=False,
        end="\n",
    ):
        selected_log_files = [
            f
            for flevel, f in log_files
            if level is LOG_ALWAYS
            or (level <= flevel and (max_level is None or flevel < max_level))
        ]
        for i in selected_log_files:
            if force_stderr and i.fileno() == 1 and not force_stdout:
                continue  # skip stdout if we write to stderr
            if force_stdout and i.fileno() == 2 and not force_stderr:
                continue  # skip stderr if we write to stdout
            i.write(str(text) + end)
            i.flush()
            if i.fileno() > 2:  # stderr
                os.fsync(i.fileno())
        for force, fno, sys_file in (
            (force_stdout, 1, sys.stdout),
            (force_stderr, 2, sys.stderr),
        ):
            if force and not any(
                i.fileno() == fno for i in selected_log_files
            ):  # not already writing to std{out,err}
                if PY3:
                    sys_file.buffer.write(text.encode("utf-8"))
                    sys_file.buffer.write(end.encode("utf-8"))
                    sys_file.flush()
                else:
                    print(text, end=end, file=sys_file)

    return log


DEFAULT_LOG = make_logger([(DEFAULT_VERBOSITY, sys.stderr)])


class DeprecatedAction(argparse.Action):
    def __init__(self, replacement=None, *args, **kwargs):
        if replacement is None:
            raise ValueError("replacement must not be None")
        super(DeprecatedAction, self).__init__(*args, **kwargs)
        self.replacement = replacement

    def __call__(self, parser, namespace, values, option_string=None):
        print(
            "ERROR: option %s is deprecated.  Use %s instead."
            % (option_string, self.replacement),
            file=sys.stderr,
        )
        sys.exit(0)


class ArgAppendWithWarningAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        items = getattr(namespace, self.dest) or []
        items.append(value)
        setattr(namespace, self.dest, items)
        if value.startswith("-w "):
            print(
                (
                    'WARNING: You seem to be trying to pass a warning list to Coq via a single %s ("%s").'
                    + "\n  I will continue anyway, but this will most likely not work."
                    + "\n  Instead try using multiple invocations, as in"
                    + "\n    %s"
                )
                % (
                    option_string,
                    value,
                    " ".join(option_string + "=" + v for v in value.split(" ")),
                ),
                file=sys.stderr,
            )


def add_libname_arguments_gen(parser, passing):
    passing_dash = passing + "-" if passing else ""
    twodash_passing_dash = "--" + passing_dash
    onedash_passing_dash = "-" + ("-" + passing + "-" if passing else "")
    passing_underscore = passing + "_" if passing else ""
    passing_for = ", for --" + passing + "coqc" if passing else ""
    parser.add_argument(
        twodash_passing_dash + "topname",
        metavar="TOPNAME",
        dest=passing_underscore + "topname",
        type=str,
        default="Top",
        action=DeprecatedAction,
        replacement="-R",
        help="The name to bind to the current directory using -R .",
    )
    parser.add_argument(
        onedash_passing_dash + "R",
        metavar=("DIR", "COQDIR"),
        dest=passing_underscore + "libnames",
        type=str,
        default=[],
        nargs=2,
        action=CoqLibnameAction,
        help="recursively map physical DIR to logical COQDIR, as in the -R argument to coqc"
        + passing_for,
    )
    parser.add_argument(
        onedash_passing_dash + "Q",
        metavar=("DIR", "COQDIR"),
        dest=passing_underscore + "non_recursive_libnames",
        type=str,
        default=[],
        nargs=2,
        action=CoqLibnameAction,
        help="(nonrecursively) map physical DIR to logical COQDIR, as in the -Q argument to coqc"
        + passing_for,
    )
    parser.add_argument(
        onedash_passing_dash + "I",
        metavar="DIR",
        dest=passing_underscore + "ocaml_dirnames",
        type=str,
        default=[],
        action="append",
        help="Look for ML files in DIR, as in the -I argument to coqc" + passing_for,
    )
    parser.add_argument(
        twodash_passing_dash + "arg",
        metavar="ARG",
        dest=passing_underscore + "coq_args",
        type=str,
        action=ArgAppendWithWarningAction,
        help='Arguments to pass to coqc and coqtop; e.g., " -indices-matter" (leading and trailing spaces are stripped)'
        + passing_for,
    )
    parser.add_argument(
        onedash_passing_dash + "f",
        metavar="FILE",
        dest=passing_underscore + "CoqProjectFile",
        nargs=1,
        type=argparse.FileType("r"),
        default=None,
        help=("A _CoqProject file" + passing_for),
    )


def add_libname_arguments(parser):
    add_libname_arguments_gen(parser, passing="")


def add_passing_libname_arguments(parser):
    add_libname_arguments_gen(parser, "passing")
    add_libname_arguments_gen(parser, "nonpassing")


def TupleType(*types):
    def call(strings):
        vals = strings.split(",")
        if len(vals) != len(types):
            raise ValueError(
                "Got %d arguments, expected %d arguments" % (len(vals), len(types))
            )
        return tuple(ty(val) for ty, val in zip(types, vals))

    return call


def add_logging_arguments(parser):
    parser.add_argument(
        "--verbose",
        "-v",
        dest="verbose",
        action="count",
        help="display some extra information by default (does not impact --verbose-log-file)",
    )
    parser.add_argument(
        "--quiet", "-q", dest="quiet", action="count", help="the inverse of --verbose"
    )
    parser.add_argument(
        "--log-file",
        "-l",
        dest="log_files",
        nargs="+",
        type=argparse.FileType("w"),
        default=[sys.stderr],
        help="The files to log output to.  Use - for stdout.",
    )
    parser.add_argument(
        "--verbose-log-file",
        dest="verbose_log_files",
        nargs="+",
        type=TupleType(int, argparse.FileType("w")),
        default=[],
        help=(
            (
                "The files to log output to at verbosity other than the default verbosity (%d if no -v/-q arguments are passed); "
                + "each argument should be a pair of a verbosity level and a file name. "
                "Use - for stdout."
            )
            % DEFAULT_VERBOSITY
        ),
    )


def process_logging_arguments(args):
    if args.verbose is None:
        args.verbose = DEFAULT_VERBOSITY
    if args.quiet is None:
        args.quiet = 0
    args.verbose -= args.quiet
    args.log = make_logger(
        [(args.verbose, f) for f in args.log_files] + args.verbose_log_files
    )
    del args.quiet
    del args.verbose
    return args


def tokenize_CoqProject(contents):
    is_in_string = False
    is_in_comment = False
    cur = ""
    for ch in contents:
        if is_in_string:
            cur += ch
            if ch == '"':
                yield cur
                cur = ""
                is_in_string = False
        elif is_in_comment:
            if ch in "\n\r":
                is_in_comment = False
        elif ch == '"':
            cur += ch
            is_in_string = True
        elif ch == "#":
            if cur:
                yield cur
            cur = ""
            is_in_comment = True
        else:
            if ch in "\n\r\t ":
                if cur:
                    yield cur
                cur = ""
            else:
                cur += ch
    if cur:
        yield cur


def argstring_to_iterable(arg):
    if arg[:1] == '"' and arg[-1:] == '"':
        arg = arg[1:-1]
    return arg.split(" ")


def append_coq_arg(env, arg, passing=""):
    for key in ("coqc_args", "coqtop_args"):
        env[passing + key] = tuple(
            list(env.get(passing + key, [])) + list(argstring_to_iterable(arg))
        )


def process_CoqProject(env, contents, passing=""):
    if contents is None:
        return
    tokens = tuple(tokenize_CoqProject(contents))
    i = 0
    while i < len(tokens):
        if tokens[i] == "-R" and i + 2 < len(tokens):
            env[passing + "libnames"].append((tokens[i + 1], tokens[i + 2]))
            i += 3
        elif tokens[i] == "-Q" and i + 2 < len(tokens):
            env[passing + "non_recursive_libnames"].append(
                (tokens[i + 1], tokens[i + 2])
            )
            i += 3
        elif tokens[i] == "-I" and i + 1 < len(tokens):
            env[passing + "ocaml_dirnames"].append(tokens[i + 1])
            i += 2
        elif tokens[i] == "-arg" and i + 1 < len(tokens):
            append_coq_arg(env, tokens[i + 1], passing=passing)
            i += 2
        elif tokens[i][-2:] == ".v":
            env[passing + "_CoqProject_v_files"].append(tokens[i])
            i += 1
        elif any(
            tokens[i][-len(ext) :] == ext
            for ext in (".mli", ".ml", ".mlg", ".mllib", ".ml4")
        ):
            i += 1
        else:
            if "log" in env.keys():
                env["log"]("Unknown _CoqProject entry: %s" % repr(tokens[i]))
            env[passing + "_CoqProject_unknown"].append(tokens[i])
            i += 1


def update_env_with_libname_default(env, args, default=((".", "Top"),), passing=""):
    if (
        len(
            getattr(args, passing + "libnames")
            + getattr(args, passing + "non_recursive_libnames")
            + env[passing + "libnames"]
            + env[passing + "non_recursive_libnames"]
        )
        == 0
    ):
        env[passing + "libnames"].extend(list(default))


def update_env_with_libnames(
    env,
    args,
    default=((".", "Top"),),
    include_passing=False,
    use_default=True,
    use_passing_default=True,
):
    all_keys = (
        "libnames",
        "non_recursive_libnames",
        "ocaml_dirnames",
        "_CoqProject",
        "_CoqProject_v_files",
        "_CoqProject_unknown",
    )
    passing_opts = ("",) if not include_passing else ("", "passing_", "nonpassing_")
    for passing in passing_opts:
        for key in all_keys:
            env[passing + key] = []
        if getattr(args, passing + "CoqProjectFile"):
            for f in getattr(args, passing + "CoqProjectFile"):
                env[passing + "_CoqProject"] = f.read()
                f.close()
        process_CoqProject(env, env[passing + "_CoqProject"], passing=passing)
        env[passing + "libnames"].extend(getattr(args, passing + "libnames"))
        env[passing + "non_recursive_libnames"].extend(
            getattr(args, passing + "non_recursive_libnames")
        )
        env[passing + "ocaml_dirnames"].extend(
            getattr(args, passing + "ocaml_dirnames")
        )
    if include_passing:
        # The nonpassing_ prefix is actually temporary; the semantics
        # are that, e.g., libnames = libnames + nonpassing_libnames,
        # while passing_libnames = libnames + passing_libnames
        for key in all_keys:
            env["passing_" + key] = env[key] + env["passing_" + key]
            env[key] = env[key] + env["nonpassing_" + key]
            del env["nonpassing_" + key]
    if use_default:
        update_env_with_libname_default(env, args, default=default)
    if use_passing_default and include_passing:
        update_env_with_libname_default(env, args, default=default, passing="passing_")


def update_env_with_coqpath_folders(passing_prefix, env, *coqpaths, skip_dirs=None):
    if skip_dirs is None:
        skip_dirs = []

    existing_names = set()
    for libnames_key in ["libnames", "non_recursive_libnames"]:
        for d, name in env.get(passing_prefix + libnames_key, []):
            if os.path.isdir(d):
                existing_names.add(name)

    def do_with_path(path):
        env.get(passing_prefix + "non_recursive_libnames", []).extend(
            (os.path.join(path, d), d)
            for d in sorted(os.listdir(path))
            if d not in skip_dirs and d not in existing_names
        )

    for coqpath in coqpaths:
        if os.path.isdir(coqpath):
            do_with_path(coqpath)
        elif ";" in coqpath:
            for path in coqpath.split(";"):
                do_with_path(path if path != "" else ".")
        elif ":" in coqpath:
            for path in coqpath.split(":"):
                do_with_path(path if path != "" else ".")


# http://stackoverflow.com/questions/5943249/python-argparse-and-controlling-overriding-the-exit-status-code
class ArgumentParser(argparse.ArgumentParser):
    def _get_action_from_name(self, name):
        """Given a name, get the Action instance registered with this parser.
        If only it were made available in the ArgumentError object. It is
        passed as it's first arg...
        """
        container = self._actions
        if name is None:
            return None
        for action in container:
            if "/".join(action.option_strings) == name:
                return action
            elif action.metavar == name:
                return action
            elif action.dest == name:
                return action

    def error(self, message):
        def reraise(extra=""):
            super(ArgumentParser, self).error(message + extra)

        exc = sys.exc_info()[1]
        if exc:
            exc.argument = self._get_action_from_name(exc.argument_name)
            exc.reraise = reraise
            raise exc
        reraise()


# returns a mapping from constant names to values to lists of command-line flags
def get_parser_name_mapping(parser):
    mapping = {}
    for action in parser._actions:
        if (
            hasattr(action, "const")
            and action.const is not None
            and len(action.option_strings) > 0
        ):
            dest = action.dest
            const = action.const
            if dest not in mapping:
                mapping[dest] = {}
            if const not in mapping[dest]:
                mapping[dest][const] = []
            mapping[dest][const] += action.option_strings
    return mapping
